pub mod sloth;

use base64::{prelude::BASE64_URL_SAFE, Engine};
use num_bigint::BigUint;
use serde::{Deserialize, Serialize};
use wasm_bindgen::prelude::*;

cfg_if::cfg_if! {
    if #[cfg(target_arch = "wasm32")] {
    } else {
        use std::collections::HashSet;
        use std::sync::{Arc, RwLock};
        use std::time::{SystemTime, UNIX_EPOCH};
        use rand::Rng;
        pub mod prime;
        pub mod ed25519;
    }
}

#[cfg(not(target_arch = "wasm32"))]
static DIFFICULTY_SCALE: f64 = 0.04;

#[derive(Serialize, Deserialize)]
pub struct Puzzle {
    pub ticket: u64,
    pub issued_at: u64,
    pub prime: BigUint,
    pub challenge: BigUint,
    pub difficulty: u32,
}

#[cfg(not(target_arch = "wasm32"))]
pub struct ChallengerBuilder {
    difficulty: Option<u32>,
    algorithm: Option<crate::sloth::Sloth>,
    signer: Option<crate::ed25519::ChallengeSigner>,
    valid_time_window: Option<u64>,
}

#[cfg(not(target_arch = "wasm32"))]
impl ChallengerBuilder {
    pub fn new() -> Self {
        Self {
            difficulty: None,
            algorithm: None,
            signer: None,
            valid_time_window: None,
        }
    }

    pub fn with_difficulty(mut self, difficulty: u32) -> Self {
        self.difficulty = Some((difficulty as f64 * DIFFICULTY_SCALE) as u32);
        self
    }

    pub fn with_prime(mut self, prime: BigUint) -> Self {
        self.algorithm = Some(crate::sloth::Sloth::new(prime));
        self
    }

    pub fn with_sign_key(mut self, sign_key: [u8; 32]) -> Self {
        self.signer = Some(crate::ed25519::ChallengeSigner::new_from_bytes(sign_key));
        self
    }

    pub fn with_valid_time_window(mut self, valid_time_window: u64) -> Self {
        self.valid_time_window = Some(valid_time_window);
        self
    }

    pub fn build(self) -> Challenger {
        let difficulty = self
            .difficulty
            .unwrap_or((1000.0 * DIFFICULTY_SCALE) as u32);

        let algorithm = self.algorithm.unwrap_or_else(|| {
            crate::sloth::Sloth::new(crate::prime::generate_prime_mod_3_4(4096, 64))
        });

        let signer = self
            .signer
            .unwrap_or_else(|| crate::ed25519::ChallengeSigner::new());

        let tickets_current = Arc::new(RwLock::new(HashSet::new()));
        let tickets_previous = Arc::new(RwLock::new(HashSet::new()));

        let valid_time_window = self.valid_time_window.unwrap_or(300);
        let last_rotation = Arc::new(RwLock::new(
            SystemTime::now()
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs(),
        ));

        Challenger {
            difficulty,
            algorithm,
            signer,
            tickets_current,
            tickets_previous,
            valid_time_window,
            last_rotation,
        }
    }
}

#[cfg(not(target_arch = "wasm32"))]
pub struct Challenger {
    pub difficulty: u32,
    algorithm: crate::sloth::Sloth,
    signer: crate::ed25519::ChallengeSigner,
    tickets_current: Arc<RwLock<HashSet<u64>>>,
    tickets_previous: Arc<RwLock<HashSet<u64>>>,
    valid_time_window: u64,
    last_rotation: Arc<RwLock<u64>>,
}

#[cfg(not(target_arch = "wasm32"))]
impl Challenger {
    pub fn set_difficulty(&mut self, difficulty: u32) {
        self.difficulty = (difficulty as f64 * DIFFICULTY_SCALE) as u32;
    }

    pub fn issue_challenge(&mut self) -> String {
        let (_x, y, p) = self.algorithm.create(self.difficulty);

        let puzzle = Puzzle {
            ticket: rand::thread_rng().gen(),
            issued_at: SystemTime::now()
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs(),
            prime: p,
            challenge: y,
            difficulty: self.difficulty,
        };

        let json = serde_json::to_string(&puzzle).unwrap();
        let json_payload = BASE64_URL_SAFE.encode(json);
        let signature = self.signer.sign(json_payload.as_bytes());
        let signature_payload = BASE64_URL_SAFE.encode(signature.to_bytes());

        let challenge = json_payload + "." + signature_payload.as_str();
        challenge
    }

    pub fn verify_challenge(&mut self, challenge: &str, solution: &str) -> bool {
        let now = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .unwrap()
            .as_secs();

        let need_rotation = {
            let last_rotation = self.last_rotation.read().unwrap();
            now - *last_rotation > self.valid_time_window
        };

        if need_rotation {
            let mut last_rotation = self.last_rotation.write().unwrap();
            let mut tickets_previous = self.tickets_previous.write().unwrap();
            let mut tickets_current = self.tickets_current.write().unwrap();
            *tickets_previous = std::mem::take(&mut *tickets_current);
            *last_rotation = now;
        }

        // Reject challenges that are too long
        if challenge.len() > 10240 || solution.len() > 1024 {
            return false;
        }

        // Split the challenge into the JSON payload and the signature
        let parts: Vec<&str> = challenge.split('.').collect();
        if parts.len() != 2 {
            return false;
        }

        let json_payload = parts[0];
        let signature_payload = parts[1];

        // Decode the signature
        let signature_bytes: [u8; 64] = match BASE64_URL_SAFE.decode(signature_payload.as_bytes()) {
            Ok(bytes) => match bytes.try_into() {
                Ok(b) => b,
                Err(_) => return false,
            },
            Err(_) => return false,
        };

        // Verify the signature on the JSON payload
        if !self
            .signer
            .verify(json_payload.as_bytes(), &signature_bytes)
        {
            return false;
        }

        // Decode the JSON payload
        let puzzle: Puzzle = match BASE64_URL_SAFE.decode(json_payload.as_bytes()) {
            Ok(bytes) => match serde_json::from_slice(&bytes) {
                Ok(s) => s,
                Err(_) => return false,
            },
            Err(_) => return false,
        };

        // Reject challenges that are too old
        if puzzle.issued_at < now - self.valid_time_window {
            return false;
        }

        // Reject challenges that have already been solved
        {
            if self
                .tickets_current
                .read()
                .unwrap()
                .contains(&puzzle.ticket)
            {
                return false;
            }
        }

        {
            if self
                .tickets_previous
                .read()
                .unwrap()
                .contains(&puzzle.ticket)
            {
                return false;
            }
        }

        // Decode the solution
        let solution: BigUint = match BASE64_URL_SAFE.decode(solution.as_bytes()) {
            Ok(bytes) => BigUint::from_bytes_be(&bytes),
            Err(_) => return false,
        };

        // Verify the solution
        let solution_result =
            self.algorithm
                .verify(&solution, &puzzle.challenge, puzzle.difficulty);

        if solution_result {
            self.tickets_current.write().unwrap().insert(puzzle.ticket);
        }

        solution_result
    }
}

#[wasm_bindgen]
pub fn solve_challenge(challenge: &str) -> String {
    let parts: Vec<&str> = challenge.split('.').collect();
    if parts.len() != 2 {
        return "ERROR".to_string();
    }

    let json_payload = parts[0];
    let puzzle: Puzzle = match BASE64_URL_SAFE.decode(json_payload.as_bytes()) {
        Ok(bytes) => match serde_json::from_slice(&bytes) {
            Ok(s) => s,
            Err(_) => return "ERROR".to_string(),
        },
        Err(_) => return "ERROR".to_string(),
    };

    let y = puzzle.challenge.clone();
    let p = puzzle.prime.clone();
    let t = puzzle.difficulty;

    let solution = crate::sloth::Sloth::decode(&y, &p, t);

    BASE64_URL_SAFE.encode(solution.to_bytes_be())
}

#[cfg(not(target_arch = "wasm32"))]
#[cfg(test)]
mod tests {
    use super::*;
    use std::thread;

    #[test]
    fn test_challenge() {
        let p = crate::prime::generate_prime_mod_3_4(4096, 64);
        println!("prime: {}", p);
        let mut challenger = ChallengerBuilder::new()
            .with_prime(p)
            .with_difficulty(1000)
            .build();
        let challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);
        println!("challenge: {}", challenge);
        println!("solution: {}", solution);
        assert!(challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_incorrect() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .build();
        let challenge = challenger.issue_challenge();
        let mut solution = solve_challenge(&challenge);
        solution.replace_range(1..4, "abcd");
        assert_eq!(false, challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_invaild_challenge() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .build();
        let mut challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);
        challenge.replace_range(1..4, "abcd");
        assert_eq!(false, challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_invaild_signature() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .build();
        let mut challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);
        challenge.replace_range(challenge.len() - 5..challenge.len() - 1, "abcd");
        assert_eq!(false, challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_reuse_attack() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .build();
        let challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);

        assert!(challenger.verify_challenge(&challenge, &solution));
        assert_eq!(false, challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_invaild_data() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .build();

        assert_eq!(
            false,
            challenger.verify_challenge(&"".to_string(), &"".to_string())
        );
        assert_eq!(
            false,
            challenger.verify_challenge(&"rin".to_string(), &"cat".to_string())
        );
    }

    #[test]
    fn test_challenge_expired() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .with_valid_time_window(1)
            .build();

        let challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);

        thread::sleep(std::time::Duration::from_secs(2));

        assert_eq!(false, challenger.verify_challenge(&challenge, &solution));
    }

    #[test]
    fn test_challenge_tickets() {
        let mut challenger = ChallengerBuilder::new()
            .with_prime(crate::prime::generate_prime_mod_3_4(2048, 64))
            .with_valid_time_window(5)
            .build();

        // Normal case
        let challenge = challenger.issue_challenge();
        let solution = solve_challenge(&challenge);

        assert_eq!(true, challenger.verify_challenge(&challenge, &solution));

        // expired case
        let challenge1 = challenger.issue_challenge();
        let solution1 = solve_challenge(&challenge1);

        thread::sleep(std::time::Duration::from_secs(6));

        // Normal case
        let challenge2 = challenger.issue_challenge();
        let solution2 = solve_challenge(&challenge2);

        assert_eq!(false, challenger.verify_challenge(&challenge1, &solution1));
        assert_eq!(true, challenger.verify_challenge(&challenge2, &solution2));
    }
}
